// Astar.cpp
// http://en.wikipedia.org/wiki/A*
// Compiler: Dev-C++ 4.9.9.2
// FB - 201012256
#include <iostream>
#include <iomanip>
#include <queue>
#include <string>
#include <math.h>
#include <ctime>

#include <stdlib.h>
#include <cstdlib>
#include <stdio.h>

#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/core/core.hpp"
//#include <wiringSerial.h>

//-------------------------------------------------------------------SERIAL
#include <termios.h>
#include <sys/fcntl.h>
#include <unistd.h>
struct termios topt;
char port[1024];
int ttyFid;
//------------------------------------------------------------------SERIAL


using namespace cv;
using namespace std;

int asdf;
Mat out;
Mat out2;

const int n=80; // horizontal size of the map
const int m=60; // vertical size size of the map
static int mapv[n][m];
static int closed_nodes_map[n][m]; // map of closed (tried-out) nodes
static int open_nodes_map[n][m]; // map of open (not-yet-tried) nodes
static int dir_map[n][m]; // map of directions
const int dir=8; // number of possible directions to go at any position
// if dir==4
//static int dx[dir]={1, 0, -1, 0};
//static int dy[dir]={0, 1, 0, -1};
// if dir==8
static int dx[dir]={1, 1, 0, -1, -1, -1, 0, 1};
static int dy[dir]={0, 1, 1, 1, 0, -1, -1, -1};

class node
{
    // current position
    int xPos;
    int yPos;
    // total distance already travelled to reach the node
    int level;
    // priority=level+remaining distance estimate
    int priority;  // smaller: higher priority

    public:
        node(int xp, int yp, int d, int p) 
            {xPos=xp; yPos=yp; level=d; priority=p;}

        int getxPos() const {return xPos;}
        int getyPos() const {return yPos;}
        int getLevel() const {return level;}
        int getPriority() const {return priority;}

        void updatePriority(const int & xDest, const int & yDest)
        {
             priority=level+estimate(xDest, yDest)*10; //A*
        }

        // give better priority to going strait instead of diagonally
        void nextLevel(const int & i) // i: direction
        {
             level+=(dir==8?(i%2==0?10:14):10);
        }

        // Estimation function for the remaining distance to the goal.
        const int & estimate(const int & xDest, const int & yDest) const
        {
            static int xd, yd, d;
            xd=xDest-xPos;
            yd=yDest-yPos;

            // Euclidian Distance
            d=static_cast<int>(sqrt(xd*xd+yd*yd));

            // Manhattan distance
            //d=abs(xd)+abs(yd);

            // Chebyshev distance
            //d=max(abs(xd), abs(yd));

            return(d);
        }
};

// Determine priority (in the priority queue)
bool operator<(const node & a, const node & b)
{
  return a.getPriority() > b.getPriority();
}

// A-star algorithm.
// The route returned is a string of direction digits.
string pathFind( const int & xStart, const int & yStart, 
                 const int & xFinish, const int & yFinish )
{
    static priority_queue<node> pq[2]; // list of open (not-yet-tried) nodes
    static int pqi; // pq index
    static node* n0;
    static node* m0;
    static int i, j, x, y, xdx, ydy;
    static char c;
    pqi=0;

    // reset the node maps
    for(y=0;y<m;y++)
    {
        for(x=0;x<n;x++)
        {
            closed_nodes_map[x][y]=0;
            open_nodes_map[x][y]=0;
        }
    }

    // create the start node and push into list of open nodes
    n0=new node(xStart, yStart, 0, 0);
    n0->updatePriority(xFinish, yFinish);
    pq[pqi].push(*n0);
    open_nodes_map[x][y]=n0->getPriority(); // mark it on the open nodes map

    // A* search
    while(!pq[pqi].empty())
    {
        // get the current node w/ the highest priority
        // from the list of open nodes
        n0=new node( pq[pqi].top().getxPos(), pq[pqi].top().getyPos(), 
                     pq[pqi].top().getLevel(), pq[pqi].top().getPriority());

        x=n0->getxPos(); y=n0->getyPos();

        pq[pqi].pop(); // remove the node from the open list
        open_nodes_map[x][y]=0;
        // mark it on the closed nodes map
        closed_nodes_map[x][y]=1;

        // quit searching when the goal state is reached
        //if((*n0).estimate(xFinish, yFinish) == 0)
        if(x==xFinish && y==yFinish) 
        {
            // generate the path from finish to start
            // by following the directions
            string path="";
            while(!(x==xStart && y==yStart))
            {
                j=dir_map[x][y];
                c='0'+(j+dir/2)%dir;
                path=c+path;
                x+=dx[j];
                y+=dy[j];
            }

            // garbage collection
            delete n0;
            // empty the leftover nodes
            while(!pq[pqi].empty()) pq[pqi].pop();
            return path;
        }

        // generate moves (child nodes) in all possible directions
        for(i=0;i<dir;i++)
        {
            xdx=x+dx[i]; ydy=y+dy[i];

            if(!(xdx<0 || xdx>n-1 || ydy<0 || ydy>m-1 || mapv[xdx][ydy]==1
                || closed_nodes_map[xdx][ydy]==1))
            {
                // generate a child node
                m0=new node( xdx, ydy, n0->getLevel(),
                             n0->getPriority());
                m0->nextLevel(i);
                m0->updatePriority(xFinish, yFinish);

                // if it is not in the open list then add into that
                if(open_nodes_map[xdx][ydy]==0)
                {
                    open_nodes_map[xdx][ydy]=m0->getPriority();
                    pq[pqi].push(*m0);
                    // mark its parent node direction
                    dir_map[xdx][ydy]=(i+dir/2)%dir;
                }
                else if(open_nodes_map[xdx][ydy]>m0->getPriority())
                {
                    // update the priority info
                    open_nodes_map[xdx][ydy]=m0->getPriority();
                    // update the parent direction info
                    dir_map[xdx][ydy]=(i+dir/2)%dir;

                    // replace the node
                    // by emptying one pq to the other one
                    // except the node to be replaced will be ignored
                    // and the new node will be pushed in instead
                    while(!(pq[pqi].top().getxPos()==xdx &&
                           pq[pqi].top().getyPos()==ydy))
                    {
                        pq[1-pqi].push(pq[pqi].top());
                        pq[pqi].pop();
                    }
                    pq[pqi].pop(); // remove the wanted node

                    // empty the larger size pq to the smaller one
                    if(pq[pqi].size()>pq[1-pqi].size()) pqi=1-pqi;
                    while(!pq[pqi].empty())
                    {
                        pq[1-pqi].push(pq[pqi].top());
                        pq[pqi].pop();
                    }
                    pqi=1-pqi;
                    pq[pqi].push(*m0); // add the better node instead
                }
                else delete m0; // garbage collection
            }
        }
        delete n0; // garbage collection
    }
    return ""; // no route found
}

//-------------------------------------------------------------------------------------------------------------------------

int element_shape = MORPH_RECT;
int max_iters = 10;
int open_close_pos = 0;
int erode_dilate_pos = 0;

static void testloop(int, void*)
{
    for(;;)
    {
        int c;
        c = waitKey(0);
        if( (char)c == 27 )
            break;
    }
}


//-----------------------------------------------------------------------------------------------------------------------------

int main()
{

    //--------------------------------------------------------------SERIAL
    ttyFid = open( "/dev/ttyUSB0", O_RDWR | O_NOCTTY | O_SYNC); // ttyUSB0 ttyAMA0
    memset( &topt, 0, sizeof(topt) );
    if(ttyFid == -1)
    {
        printf( "Unable to open port keith\n" );
        return -1;
    }

    cfsetispeed( &topt, B115200);
    cfsetospeed( &topt, B115200);

    topt.c_cflag = (topt.c_cflag & ~CSIZE) | CS8;
    topt.c_iflag &= ~IGNBRK;
    topt.c_lflag = 0;
    topt.c_oflag = 0;
    topt.c_cc[VMIN] = 0;
    topt.c_cc[VTIME] = 5;
    topt.c_iflag &= ~(IXON | IXOFF | IXANY );
    topt.c_cflag |= (CLOCAL | CREAD);
    topt.c_cflag &= ~(PARENB | PARODD);
    topt.c_cflag |= 0; //parity
    topt.c_cflag &= ~CSTOPB;
    //topt.c_cflag &= ~CSIZE;
    //topt.c_cflag |= CS8;
    topt.c_cflag &= ~CRTSCTS;

    if( tcgetattr( ttyFid, &topt ) != 0 );
    {
        printf("Couldn't get Term Attributes keith\n");
        //return -1;
    }

    printf("BEFORE TCGETATTR\n");
    tcgetattr( ttyFid, &topt );
    printf("AFTER TCGETATTR\n");
    tcsetattr( ttyFid, TCSANOW, &topt );
    if( tcsetattr( ttyFid, TCSAFLUSH, &topt ) < 0 )
    {
        printf("Set Term Att Fail keith\n");
        return -1;
    }
    char ch;
    ch = 10;
    //while(1)
    //{
    write( ttyFid, &ch, sizeof(ch) );
    //write( ttyFid, "%d", ch );
    //}
    //--------------------------------------------------------------SERIAL

    int bhl = 0;
    int bhh = 5;
    int bsl = 128;
    int bsh = 255;
    int bvl = 128;
    int bvh = 255;

    int rhl = 90;
    int rhh = 120;
    int rsl = 0;
    int rsh = 255;
    int rvl = 0;
    int rvh = 255;

    srand(time(NULL));
    //Mat img = imread( "simplesm.jpg", CV_LOAD_IMAGE_COLOR );
    Mat img = imread( "IMG_6050med.jpg", CV_LOAD_IMAGE_COLOR );
    Mat img2;
    out = cvCreateImage( img.size(), 8, 1 );
    out2 = cvCreateImage( img.size(), 8, 1 );
    namedWindow( "win1", 0);
    namedWindow( "win2", 0);
    namedWindow( "win3", 0);
    imshow( "win1", img );
    cvtColor(img, img, CV_BGR2HSV );

    Scalar blb(bhl,bsl,bvl,0); //lower bound for blanket
    Scalar bub(bhh,bsh,bvh,0); //upper bound for blanket
    inRange( img, blb, bub, out ); //find blanket
    Mat element = getStructuringElement( MORPH_RECT, Size(7,7), Point(3,3) );
    erode( out, out, element);
    element = getStructuringElement( MORPH_RECT, Size(41,41), Point(20,20) );
    dilate( out, out, element);
    erode( out, out, element); //blanket mask is ready

    //img.copyTo( img2, out);
    bitwise_and(img, img, img2, out); // src1, src2, dst, mask

    Scalar rlb(rhl,rsl,rvl,0);
    Scalar rub(rhh,rsh,rvh,0);
    inRange( img2, rlb, rub, img2 );

    element = getStructuringElement( MORPH_RECT, Size(5,5), Point(2,2) );
    erode( img2, img2, element);
    element = getStructuringElement( MORPH_RECT, Size(11,11), Point(5,5) );
    dilate( img2, img2, element);
    element = getStructuringElement( MORPH_RECT, Size(5,5), Point(2,2) );
    erode( img2, img2, element);

    out = 255 - out;
    out2 = img2 + out;

    resize( out2, out2, Size(80,60), 0, 0, INTER_LINEAR );

    imshow( "win2", out );
    imshow( "win3", out2 );
    //createTrackbar("it", "win3",&erode_dilate_pos,max_iters*2+1,Dilate);
    cvWaitKey(0);

/*
    // create empty map
    for(int y=0;y<m;y++)
    {
        for(int x=0;x<n;x++) mapv[x][y]=0;
    }

    // fillout the map matrix with a '+' pattern
    for(int x=n/8;x<n*7/8;x++)
    {
        mapv[x][m/2]=1;
    }
    for(int y=m/8;y<m*7/8;y++)
    {
        mapv[n/2][y]=1;
    }
*/

    for ( int xx = 0; xx < 80; xx++ )
    {
        for (int yy = 0; yy < 60; yy++)
        {
            asdf = out2.at<uchar>(yy,xx);
            mapv[xx][yy] = asdf/255;
            cout << mapv[xx][yy];
        }
    }

    int xA,yA,xB,yB;
    xB = 35;
    yB = 30;
    xA = 55;
    yA = 45;
/*
    // randomly select start and finish locations
    int xA, yA, xB, yB;
    switch(rand()%8)
    {
        case 0: xA=0;yA=0;xB=n-1;yB=m-1; break;
        case 1: xA=0;yA=m-1;xB=n-1;yB=0; break;
        case 2: xA=n/2-1;yA=m/2-1;xB=n/2+1;yB=m/2+1; break;
        case 3: xA=n/2-1;yA=m/2+1;xB=n/2+1;yB=m/2-1; break;
        case 4: xA=n/2-1;yA=0;xB=n/2+1;yB=m-1; break;
        case 5: xA=n/2+1;yA=m-1;xB=n/2-1;yB=0; break;
        case 6: xA=0;yA=m/2-1;xB=n-1;yB=m/2+1; break;
        case 7: xA=n-1;yA=m/2+1;xB=0;yB=m/2-1; break;
    }
*/
    cout<<"Map Size (X,Y): "<<n<<","<<m<<endl;
    cout<<"Start: "<<xA<<","<<yA<<endl;
    cout<<"Finish: "<<xB<<","<<yB<<endl;
    // get the route
    clock_t start = clock();
    string route=pathFind(xA, yA, xB, yB);
    if(route=="") cout<<"An empty route generated!"<<endl;
    clock_t end = clock();
    double time_elapsed = double(end - start);
    cout<<"Time to calculate the route (ms): "<<time_elapsed<<endl;
    cout<<"Route:"<<endl;
    cout<<route<<endl<<endl;

    // follow the route on the map and display it
    if(route.length()>0)
    {
        int j; char c;
        int x=xA;
        int y=yA;
        mapv[x][y]=2;
        for(int i=0;i<route.length();i++)
        {
            c =route.at(i);
            j=atoi(&c);
            x=x+dx[j];
            y=y+dy[j];
            mapv[x][y]=3;
        }
        mapv[x][y]=4;

        // display the map with the route
        for(int y=0;y<m;y++)
        {
            for(int x=0;x<n;x++)
                if(mapv[x][y]==0)
                    cout<<".";
                else if(mapv[x][y]==1)
                    cout<<"O"; //obstacle
                else if(mapv[x][y]==2)
                    cout<<"S"; //start
                else if(mapv[x][y]==3)
                {
                    cout<<"R"; //route
                    //cvSet2D(img,y,x,Scalar(128,128,128,0));
                    img.at<Vec3b>(y,x)[0] = img.at<Vec3b>(y,x)[1] = img.at<Vec3b>(y,x)[2] =  128;
                }
                else if(mapv[x][y]==4)
                    cout<<"F"; //finish
            cout<<endl;
        }
    cvNamedWindow( "win", 0);
    imshow( "win", img );
    cvWaitKey(0);
    img.release();
    out.release();
    img2.release();
    out2.release();
    destroyAllWindows();
    cvWaitKey(0);
    }
    getchar(); // wait for a (Enter) keypress
    close( ttyFid );
    return(0);
}
